package ITL_eSSP;

/**
 *
 * @author Jan Pichl
 */
import java.io.*;
import java.util.*;
import javax.crypto.*;
import javax.crypto.spec.*;

//select javacomm or RxTx library
//import javax.comm.*;
import gnu.io.*;

public class SSPComms implements SerialPortEventListener {

    public final int ssp_CMD_SYNC = 0x11;
    public final int SSP_STX = 0x7F;
    public final int SSP_STEX = 0x7E;
    public final int CRC_SSP_SEED = 0xFFFF;
    public final int CRC_SSP_POLY = 0x8005;
    public final long MAX_RANDOM_INTEGER = 2147483647L;
    SerialPort comPort;
    SSP_TX_RX_PACKET ssp = new SSP_TX_RX_PACKET();
    RandomNumber rand = new RandomNumber();
    InputStream inputStream;
    OutputStream outputStream;

    public boolean OpenSSPComPort(SSP_COMMAND cmd) {
        Enumeration portList;

        CommPortIdentifier portId = null;
        portList = CommPortIdentifier.getPortIdentifiers();

        while (portList.hasMoreElements()) {
            CommPortIdentifier portIdTmp = (CommPortIdentifier) portList.nextElement();
            if (portIdTmp.getPortType() == CommPortIdentifier.PORT_SERIAL) {
                if (portIdTmp.getName().equals(cmd.ComPort)) {
                    portId = portIdTmp;
                    break;
                }
            }
        }
        if (portId == null) {
            return false;
        }

        try {
            comPort = (SerialPort) portId.open("ITL_SSP", 1000);
            comPort.setSerialPortParams(cmd.BaudRate, SerialPort.DATABITS_8, SerialPort.STOPBITS_2, SerialPort.PARITY_NONE);
            inputStream = comPort.getInputStream();
            outputStream = comPort.getOutputStream();
            comPort.addEventListener(this);
            comPort.notifyOnDataAvailable(true);

        } catch (Exception e) {
            return false;
        }

        return true;
    }

    public boolean CloseComPort() {
        try {
            if (inputStream != null) {
                inputStream.close();
            }
            if (outputStream != null) {
                outputStream.close();
            }
            if (comPort != null) {
                comPort.close();
            }
        } catch (Exception ex) {
            return false;
        }

        return true;
    }

    public boolean SSPSendCommand(SSP_COMMAND cmd, SSP_COMMAND_INFO sspInfo) {
        int i;
        byte encryptLength;
        int crcR;
        int[] tData = new int[255];
        int retry;
        long slaveCount;
        long txTime;

        /* compile the SSP packet and check for errors  */
        if (!CompileSSPCommand(cmd, sspInfo)) {
            cmd.ResponseStatus = PORT_STATUS.SSP_PACKET_ERROR;
            return false;
        }

        retry = cmd.RetryLevel;
        /* transmit the packet    */
        do {
            ssp.NewResponse = false;  /* set flag to wait for a new reply from slave   */

            ssp.rxBufferLength = 0;
            ssp.rxPtr = 0;
            if (WritePort() != true) {
                cmd.ResponseStatus = PORT_STATUS.PORT_ERROR;
                return false;
            }

            /* wait for out reply   */
            cmd.ResponseStatus = PORT_STATUS.SSP_REPLY_OK;
            txTime = System.currentTimeMillis();
            while (!ssp.NewResponse) {
                /* check for reply timeout   */
                try {
                    Thread.sleep(2);
                } catch (Exception e) {

                }

                if (System.currentTimeMillis() - txTime > cmd.Timeout) {
                    cmd.ResponseStatus = PORT_STATUS.SSP_CMD_TIMEOUT;
                    break;
                }
            }

            if (cmd.ResponseStatus == PORT_STATUS.SSP_REPLY_OK) {
                break;
            }

            retry--;
        } while (retry > 0);

        sspInfo.Receive.packetTime = (int) (System.currentTimeMillis() - txTime);

        if (cmd.ResponseStatus == PORT_STATUS.SSP_CMD_TIMEOUT) {
            sspInfo.Receive.PacketLength = 0;
            return false;
        }

        sspInfo.Receive.PacketLength = (byte) (ssp.rxData[2] + 5);
        for (i = 0; i < sspInfo.Receive.PacketLength; i++) {
            sspInfo.Receive.PacketData[i] = ssp.rxData[i];
        }

        /* load the command structure with ssp packet data   */
        if (ssp.rxData[3] == SSP_STEX) {   /* check for encrpted packet    */

            encryptLength = (byte) (ssp.rxData[2] - 1);
            //decrypt packet
            aes_decrypt(cmd.Key, ssp.rxData, encryptLength, 4);
            /* check the checksum    */
            crcR = cal_crc_loop_CCITT_A((encryptLength - 2), 4, ssp.rxData, CRC_SSP_SEED, CRC_SSP_POLY);
            if ((crcR & 0xFF) != ssp.rxData[ssp.rxData[2] + 1] || ((crcR >> 8) & 0xFF) != ssp.rxData[ssp.rxData[2] + 2]) {
                cmd.ResponseStatus = PORT_STATUS.SSP_PACKET_ERROR;
                sspInfo.Receive.PacketLength = 0;
                return false;
            }
            /* check the slave count against the host count  */
            slaveCount = 0;
            for (i = 0; i < 4; i++) {
                slaveCount += (long) (ssp.rxData[5 + i]) << (i * 8);
            }
            /* no match then we discard this packet and do not act on it's info  */
            if (slaveCount != cmd.encPktCount) {
                cmd.ResponseStatus = PORT_STATUS.SSP_PACKET_ERROR;
                sspInfo.Receive.PacketLength = 0;
                return false;
            }

            /* restore data for correct decode  */
            ssp.rxBufferLength = (byte) (ssp.rxData[4] + 5);
            tData[0] = ssp.rxData[0];
            tData[1] = ssp.rxData[1];
            tData[2] = ssp.rxData[4];
            for (i = 0; i < ssp.rxData[4]; i++) {
                tData[3 + i] = ssp.rxData[9 + i];
            }
            crcR = cal_crc_loop_CCITT_A((ssp.rxBufferLength - 3), 1, tData, CRC_SSP_SEED, CRC_SSP_POLY);
            tData[3 + ssp.rxData[4]] = (byte) (crcR & 0xFF);
            tData[4 + ssp.rxData[4]] = (byte) ((crcR >> 8) & 0xFF);
            for (i = 0; i < ssp.rxBufferLength; i++) {
                ssp.rxData[i] = tData[i];
            }
        }

        cmd.ResponseDataLength = ssp.rxData[2];
        for (i = 0; i < cmd.ResponseDataLength; i++) {
            cmd.ResponseData[i] = ssp.rxData[i + 3];
        }

        sspInfo.PreEncryptedRecieve.PacketLength = ssp.rxBufferLength;
        for (i = 0; i < ssp.rxBufferLength; i++) {
            sspInfo.PreEncryptedRecieve.PacketData[i] = ssp.rxData[i];
        }

        /* alternate the seq bit   */
        if (cmd.sspSeq == 0x80) {
            cmd.sspSeq = 0;
        } else {
            cmd.sspSeq = 0x80;
        }

        /* terminate the thread function   */
        cmd.ResponseStatus = PORT_STATUS.SSP_REPLY_OK;

        return true;
    }

    boolean CompileSSPCommand(SSP_COMMAND cmd, SSP_COMMAND_INFO sspInfo) {
        int i, j;
        int crc;
        int[] tBuffer = new int[255];

        //clear the receive buffer
        ssp.rxPtr = 0;
        for (i = 0; i < 255; i++) {
            ssp.rxData[i] = 0x00;
        }

        /* for sync commands reset the deq bit   */
        if (cmd.CommandData[0] == ssp_CMD_SYNC) {
            cmd.sspSeq = 0x80;
        }

        /* update the log packet data before any encryption   */
        sspInfo.Encrypted = cmd.EncryptionStatus;
        sspInfo.PreEncryptedTransmit.PacketLength = (byte) (cmd.CommandDataLength + 5);
        sspInfo.PreEncryptedTransmit.PacketData[0] = SSP_STX;					/* ssp packet start   */

        sspInfo.PreEncryptedTransmit.PacketData[1] = (byte) (cmd.SSPAddress | cmd.sspSeq);  /* the address/seq bit */

        sspInfo.PreEncryptedTransmit.PacketData[2] = cmd.CommandDataLength;    /* the data length only (always > 0)  */

        for (i = 0; i < cmd.CommandDataLength; i++) /* add the command data  */ {
            sspInfo.PreEncryptedTransmit.PacketData[3 + i] = cmd.CommandData[i];
        }
        /* calc the packet CRC  (all bytes except STX)   */
        crc = cal_crc_loop_CCITT_A((cmd.CommandDataLength + 2), 1, sspInfo.PreEncryptedTransmit.PacketData, CRC_SSP_SEED, CRC_SSP_POLY);
        sspInfo.PreEncryptedTransmit.PacketData[3 + cmd.CommandDataLength] = (byte) (crc & 0xFF);
        sspInfo.PreEncryptedTransmit.PacketData[4 + cmd.CommandDataLength] = (byte) ((crc >> 8) & 0xFF);

        /* is this a encrypted packet  */
        if (cmd.EncryptionStatus) {
            EncryptSSPPacket(cmd);
        }

        /* create the packet from this data   */
        ssp.CheckStuff = 0;
        ssp.SSPAddress = cmd.SSPAddress;
        ssp.rxPtr = 0;
        ssp.txPtr = 0;
        ssp.txBufferLength = (cmd.CommandDataLength + 5);  /* the full ssp packet length   */

        ssp.txData[0] = SSP_STX;					/* ssp packet start   */

        ssp.txData[1] = (cmd.SSPAddress | cmd.sspSeq);  /* the address/seq bit */

        ssp.txData[2] = cmd.CommandDataLength;    /* the data length only (always > 0)  */

        for (i = 0; i < cmd.CommandDataLength; i++) /* add the command data  */ {
            ssp.txData[3 + i] = cmd.CommandData[i];
        }
        /* calc the packet CRC  (all bytes except STX)   */
        crc = cal_crc_loop_CCITT_A((ssp.txBufferLength - 3), 1, ssp.txData, CRC_SSP_SEED, CRC_SSP_POLY);
        ssp.txData[3 + cmd.CommandDataLength] = (crc & 0xFF);
        ssp.txData[4 + cmd.CommandDataLength] = ((crc >> 8) & 0xFF);

        for (i = 0; i < ssp.txBufferLength; i++) {
            sspInfo.Transmit.PacketData[i] = ssp.txData[i];
        }
        sspInfo.Transmit.PacketLength = ssp.txBufferLength;


        /* we now need to 'byte stuff' this buffered data   */
        j = 0;
        tBuffer[j++] = ssp.txData[0];
        for (i = 1; i < ssp.txBufferLength; i++) {
            tBuffer[j] = ssp.txData[i];
            if (ssp.txData[i] == SSP_STX) {
                tBuffer[++j] = SSP_STX;   /* SSP_STX found in data so add another to 'stuff it'  */

            }
            j++;
        }
        for (i = 0; i < j; i++) {
            ssp.txData[i] = tBuffer[i];
        }
        ssp.txBufferLength = (byte) j;

        return true;
    }

    void EncryptSSPPacket(SSP_COMMAND cmd) {
        final byte FIXED_PACKET_LENGTH = 7;
        final byte C_MAX_KEY_LENGTH = 16;
        byte pkLength, i, packLength = 0;
        int crc;
        int[] tmpData = new int[255];

        pkLength = (byte) (cmd.CommandDataLength + FIXED_PACKET_LENGTH);

        /* find the length of packing data required */
        if (pkLength % C_MAX_KEY_LENGTH != 0) {
            packLength = (byte) (C_MAX_KEY_LENGTH - (pkLength % C_MAX_KEY_LENGTH));
        }
        pkLength += packLength;

        tmpData[0] = cmd.CommandDataLength; /* the length of the data without packing */

        /* add in the encrypted packet count   */
        for (i = 0; i < 4; i++) {
            tmpData[1 + i] = (byte) ((cmd.encPktCount >> (8 * i) & 0xFF));
        }

        for (i = 0; i < cmd.CommandDataLength; i++) {
            tmpData[i + 5] = cmd.CommandData[i];
        }

        /* add random packing data  */
        for (i = 0; i < packLength; i++) {
            tmpData[5 + cmd.CommandDataLength + i] = (byte)rand.GenerateRandomNumber();
        }
        /* add CRC to packet end   */

        crc = cal_crc_loop_CCITT_A((pkLength - 2), 0, tmpData, CRC_SSP_SEED, CRC_SSP_POLY);

        tmpData[pkLength - 2] = (crc & 0xFF);
        tmpData[pkLength - 1] = ((crc >> 8) & 0xFF);

        aes_encrypt(cmd.Key, tmpData, pkLength, 0);

        pkLength++; /* increment as the final length will have an STEX command added   */

        cmd.CommandDataLength = pkLength;
        cmd.CommandData[0] = SSP_STEX;
        for (i = 0; i < pkLength - 1; i++) {
            cmd.CommandData[1 + i] = tmpData[i];
        }

        cmd.encPktCount++;  /* increment the counter after a successful encrypted packet   */

    }

    void aes_encrypt(SSP_FULL_KEY sspKey, int[] data, byte length, int offset) {
        int i;
        byte[] data_byte = new byte[length];
        byte[] key = new byte[16];
        byte[] enc_data_byte;

        for (i = 0; i < length; i++) {
            data_byte[i] = (byte) data[i];
        }

        for (i = 0; i < 8; i++) {
            key[i] = (byte) (sspKey.FixedKey >> (8 * i));
            key[i + 8] = (byte) (sspKey.VariableKey >> (8 * i));
        }

        try {
            Cipher cipher = Cipher.getInstance("AES/ECB/NoPadding");
            cipher.init(Cipher.ENCRYPT_MODE, new SecretKeySpec(key, "AES"));
            ByteArrayInputStream bIn = new ByteArrayInputStream(data_byte);
            CipherInputStream cIn = new CipherInputStream(bIn, cipher);
            ByteArrayOutputStream bOut = new ByteArrayOutputStream();
            int ch;
            while ((ch = cIn.read()) >= 0) {
                bOut.write(ch);
            }
            enc_data_byte = bOut.toByteArray();
        } catch (Exception e) {
            return;
        }

        for (i = 0; i < length; i++) {
            data[i] = enc_data_byte[i];
            data[i] &= 0xFF;
        }
    }

    void aes_decrypt(SSP_FULL_KEY sspKey, int[] data, byte length, int offset) {
        int i;
        byte[] data_byte = new byte[length];
        byte[] key = new byte[16];
        byte[] enc_data_byte;

        for (i = 0; i < length; i++) {
            data_byte[i] = (byte) data[i + offset];
        }

        for (i = 0; i < 8; i++) {
            key[i] = (byte) (sspKey.FixedKey >> (8 * i));
            key[i + 8] = (byte) (sspKey.VariableKey >> (8 * i));
        }

        try {
            Cipher cipher = Cipher.getInstance("AES/ECB/NoPadding");
            cipher.init(Cipher.DECRYPT_MODE, new SecretKeySpec(key, "AES"));
            ByteArrayInputStream bIn = new ByteArrayInputStream(data_byte);
            CipherInputStream cIn = new CipherInputStream(bIn, cipher);
            ByteArrayOutputStream bOut = new ByteArrayOutputStream();
            int ch;
            while ((ch = cIn.read()) >= 0) {
                bOut.write(ch);
            }
            enc_data_byte = bOut.toByteArray();
        } catch (Exception e) {
            return;
        }

        for (i = 0; i < length; i++) {
            data[i + offset] = enc_data_byte[i];
            data[i + offset] &= 0xFF;
        }
    }

    public boolean InitiateSSPHostKeys(SSP_KEYS keys, SSP_COMMAND cmd) {
        long swap = 0;

        /* create the two random prime numbers  */
        keys.Generator = rand.GeneratePrime();
        keys.Modulus = rand.GeneratePrime();
        /* make sure Generator is larger than Modulus   */
        if (keys.Generator < keys.Modulus) {
            swap = keys.Generator;
            keys.Generator = keys.Modulus;
            keys.Modulus = swap;
        }

        if (CreateHostInterKey(keys) == false) {
            return false;
        }

        // reset the packet counter here for a successful key neg 
        cmd.encPktCount = 0;

        return true;
    }

    // Creates a host intermediate key 
    public boolean CreateHostInterKey(SSP_KEYS keys) {
        if (keys.Generator == 0 || keys.Modulus == 0) {
            return false;
        }

        keys.HostRandom = (long) (rand.GenerateRandomNumber() % MAX_RANDOM_INTEGER);
        keys.HostInter = rand.XpowYmodN(keys.Generator, keys.HostRandom, keys.Modulus);

        return true;
    }

    // creates the host encryption key   
    public boolean CreateSSPHostEncryptionKey(SSP_KEYS keys) {
        keys.KeyHost = rand.XpowYmodN(keys.SlaveInterKey, keys.HostRandom, keys.Modulus);

        return true;
    }

    private boolean WritePort() {
        try {
            int i;
            byte[] tmp = new byte[ssp.txBufferLength];

            for (i = 0; i < ssp.txBufferLength; i++) {
                tmp[i] = (byte) (ssp.txData[i] & 0xFF);
            }
            outputStream.write(tmp);
        } catch (Exception ex) {
            return false;
        }
        return true;
    }

    void SSPDataIn(int RxChar) {
        int crc;

        if (RxChar == SSP_STX && ssp.rxPtr == 0) {
            // packet start
            ssp.rxData[ssp.rxPtr++] = RxChar;
        } else {
            // if last byte was start byte, and next is not then
            // restart the packet
            if (ssp.CheckStuff == 1) {
                if (RxChar != SSP_STX) {
                    ssp.rxData[0] = SSP_STX;
                    ssp.rxData[1] = RxChar;
                    ssp.rxPtr = 2;
                } else {
                    ssp.rxData[ssp.rxPtr++] = RxChar;
                }
                // reset stuff check flag	
                ssp.CheckStuff = 0;
            } else {
                // set flag for stuffed byte check
                if (RxChar == SSP_STX) {
                    ssp.CheckStuff = 1;
                } else {
                    // add data to packet
                    ssp.rxData[ssp.rxPtr++] = RxChar;
                    // get the packet length
                    if (ssp.rxPtr == 3) {
                        ssp.rxBufferLength = (ssp.rxData[2] + 5);
                    }
                }
            }
            // are we at the end of the packet
            if (ssp.rxBufferLength > 2 && ssp.rxPtr == ssp.rxBufferLength) {
                // is this packet for us ??
                if ((ssp.rxData[1] & SSP_STX) == ssp.SSPAddress) {
                    // is the checksum correct
                    crc = cal_crc_loop_CCITT_A((ssp.rxBufferLength - 3), 1, ssp.rxData, CRC_SSP_SEED, CRC_SSP_POLY);
                    if ((crc & 0xFF) == ssp.rxData[ssp.rxBufferLength - 2] && ((crc >> 8) & 0xFF) == ssp.rxData[ssp.rxBufferLength - 1]) {
                        ssp.NewResponse = true;  /* we have a new response so set flag  */
                    }
                }
                // reset packet 
                ssp.rxPtr = 0;
                ssp.CheckStuff = 0;
                ssp.rxBufferLength = 0;
            }
        }
    }

    int cal_crc_loop_CCITT_A(int l, int offset, int[] p, int seed, int cd) {
        int i, j;
        int crc = seed;

        for (i = 0; i < l; ++i) {
            for (j = 0; j < 8; j++) {
                boolean bit = ((p[i + offset] >> (7 - j) & 1) == 1);
                boolean c15 = ((crc >> 15 & 1) == 1);
                crc <<= 1;
                if (c15 ^ bit) {
                    crc ^= cd;
                }
            }
        }
        crc &= 0xffff;
        return crc;
    }

    public void serialEvent(SerialPortEvent event) {
        try {
            while (inputStream.available() > 0) {
                SSPDataIn(inputStream.read());
            }
        } catch (Exception ex) {
        }
    }
}
